/************************************************************************/
/* protocols.c							 */
/************************************************************************/
/*
 *  (c)2021 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <errno.h>
#include <pnetbuffer.h>
#include <protocols.h>





uint8_t pnetbuffer_read_compare(pnetbuffer_t *nb, compare_t *compare)
{
	uint8_t		rc;
	rc = pnetbuffer_read_uint16(nb, &compare->action);
	if (rc) return rc;
	rc = pnetbuffer_read_uint16(nb, &compare->protocol);
	if (rc) return rc;
	rc = pnetbuffer_read_uint32(nb, &compare->compare);
	if (rc) return rc;
	rc = pnetbuffer_read_uint32(nb, &compare->src_addr.net);
	if (rc) return rc;
	rc = pnetbuffer_read_uint32(nb, &compare->src_addr.mask);
	if (rc) return rc;
	rc = pnetbuffer_read_uint32(nb, &compare->dest_addr.net);
	if (rc) return rc;
	rc = pnetbuffer_read_uint32(nb, &compare->dest_addr.mask);
	if (rc) return rc;
	rc = pnetbuffer_read_uint16(nb, &compare->src_port);
	if (rc) return rc;
	rc = pnetbuffer_read_uint16(nb, &compare->dest_port);
	if (rc) return rc;
	return 0;
}


uint8_t pnetbuffer_write_compare(pnetbuffer_t *nb, compare_t *compare)
{
	uint8_t		rc;
	rc = pnetbuffer_add_uint16(nb, compare->action);
	if (rc) return rc;
	rc = pnetbuffer_add_uint16(nb, compare->protocol);
	if (rc) return rc;
	rc = pnetbuffer_add_uint32(nb, compare->compare);
	if (rc) return rc;
	rc = pnetbuffer_add_uint32(nb, compare->src_addr.net);
	if (rc) return rc;
	rc = pnetbuffer_add_uint32(nb, compare->src_addr.mask);
	if (rc) return rc;
	rc = pnetbuffer_add_uint32(nb, compare->dest_addr.net);
	if (rc) return rc;
	rc = pnetbuffer_add_uint32(nb, compare->dest_addr.mask);
	if (rc) return rc;
	rc = pnetbuffer_add_uint16(nb, compare->src_port);
	if (rc) return rc;
	rc = pnetbuffer_add_uint16(nb, compare->dest_port);
	if (rc) return rc;
	return 0;
}






#define GENEVE_GET_VERSION(x)	((x&0xc0000000)>>30)
#define GENEVE_GET_OPTLEN(x)	((x&0x3f000000)>>24)	// return bytes not words
#define GENEVE_GET_ETHERTYPE(x)	(x&0x0000ffff)
#define GENEVE_GET_VNI(x)	((x&0xffffff00)>>8)

uint8_t pnetbuffer_read_geneve(pnetbuffer_t *nb,
				uint16_t *ethertype,
				uint32_t *vni,
				char *option_data,
				uint16_t *option_len)
{
	int		len;
	uint8_t		rc;
	uint32_t	u32;
	char		buffer[256];
	rc = pnetbuffer_read_uint32(nb, &u32);
	if (rc) return rc;
	*ethertype = GENEVE_GET_ETHERTYPE(u32);
	len = GENEVE_GET_OPTLEN(u32);
	*option_len = (uint16_t)len;

	rc = pnetbuffer_read_uint32(nb, &u32);
	if (rc) return rc;
	*vni = GENEVE_GET_VNI(u32);
	if (len > 0) {
		if (option_data) {
			rc = pnetbuffer_read_data(nb,
					option_data,
					(uint16_t)len);
			if (rc) return rc;
		} else {
			rc = pnetbuffer_read_data(nb,
					buffer,
					(uint16_t)len);
			if (rc) return rc;
		}
	}
	return 0;
}



uint8_t pnetbuffer_read_ethernet2(pnetbuffer_t *nb,
				char src_mac[6],
				char dest_mac[6],
				uint16_t *type)
{
	uint8_t		rc;
	rc = pnetbuffer_read_data(nb, dest_mac, 6);
	if (rc) return rc;
	rc = pnetbuffer_read_data(nb, src_mac, 6);
	if (rc) return rc;
	rc = pnetbuffer_read_uint16(nb, type);
	if (rc) return rc;

	return 0;
}



#define IPV4_GET_VERSION(x)	((x & 0xf0000000)>>28)
#define IPV4_GET_IHL(x)		(((x & 0x0f000000)>>24)<<2)
#define IPV4_GET_TOS(x)		((x & 0x00ff0000)>>16)
#define IPV4_GET_TOTALLEN(x)	(x & 0x0000ffff)
#define IPV4_GET_IDENT(x)	((x & 0xffff0000)>>16)
#define IPV4_GET_FLAGS(x)	((x & 0x0000e000)>>13)
#define IPV4_GET_FRAGOSET(x)	(x & 0x00001fff)
#define IPV4_GET_TTL(x)		((x & 0xff000000)>>24)
#define IPV4_GET_PROTOCOL(x)	((x & 0x00ff0000)>>16)
#define IPV4_GET_CHECKSUM(x)	(x & 0x0000ffff)

uint8_t pnetbuffer_read_ipv4(pnetbuffer_t *nb,
				uint32_t *src_ip,
				uint32_t *dest_ip,
				uint16_t *protocol,
				uint16_t *length,
				uint16_t *checksum)
{
	uint8_t		rc;
	uint8_t		ihl;
	uint32_t	u32;
	char		buffer[256];
	rc = pnetbuffer_read_uint32(nb, &u32);
	if (rc) return rc;
	if (IPV4_GET_VERSION(u32) != 4) return EPROTO;
	ihl = IPV4_GET_IHL(u32);
	*length = IPV4_GET_TOTALLEN(u32);

	rc = pnetbuffer_read_uint32(nb, &u32);
	if (rc) return rc;

	rc = pnetbuffer_read_uint32(nb, &u32);
	if (rc) return rc;
	*protocol = IPV4_GET_PROTOCOL(u32);
	*checksum = IPV4_GET_CHECKSUM(u32);

	rc = pnetbuffer_read_uint32(nb, src_ip);
	if (rc) return rc;

	rc = pnetbuffer_read_uint32(nb, dest_ip);
	if (rc) return rc;

	if (ihl > 20) {
		rc = pnetbuffer_read_data(nb, buffer, ihl-20);
		if (rc) return rc;
	}
	return 0;
}


#define ICMP_GET_TYPE(x)	((x & 0xff000000)>>16)
#define ICMP_GET_CODE(x)	((x & 0x00ff0000)>>16)
#define ICMP_GET_CHECKSUM(x)	(x & 0x0000ffff)

uint8_t pnetbuffer_read_icmp(pnetbuffer_t *nb,
				uint8_t *type,
				uint8_t *code,
				uint16_t *checksum)
{
	uint8_t		rc;
	uint32_t	u32;
	rc = pnetbuffer_read_uint32(nb, &u32);
	if (rc) return rc;
	*type = ICMP_GET_TYPE(u32);
	*code = ICMP_GET_CODE(u32);
	*checksum = ICMP_GET_CHECKSUM(u32);

	rc = pnetbuffer_read_uint32(nb, &u32);
	if (rc) return rc;

	return 0;
}



#define UDP_GET_SRC_PORT(x)	((x & 0xffff0000)>>16)
#define UDP_GET_DEST_PORT(x)	(x & 0x0000ffff)
#define UDP_GET_LENGTH(x)	((x & 0xffff0000)>>16)
#define UDP_GET_CHECKSUM(x)	(x & 0x0000ffff)

uint8_t pnetbuffer_read_udp(pnetbuffer_t *nb,
				uint16_t *src_port,
				uint16_t *dest_port,
				uint16_t *checksum,
				uint16_t *length)
{
	uint8_t		rc;
	uint32_t	u32;
	rc = pnetbuffer_read_uint32(nb, &u32);
	if (rc) return rc;
	*src_port = UDP_GET_SRC_PORT(u32);
	*dest_port = UDP_GET_DEST_PORT(u32);

	rc = pnetbuffer_read_uint32(nb, &u32);
	if (rc) return rc;
	*length = UDP_GET_LENGTH(u32);
	*checksum = UDP_GET_CHECKSUM(u32);

	return 0;
}


#define TCP_GET_SRC_PORT(x)	((x & 0xffff0000)>>16)
#define TCP_GET_DEST_PORT(x)	(x & 0x0000ffff)
#define TCP_GET_DATA_OSET(x)	((x & 0xf0000000)>>22) //  return bytes (<<2)
#define TCP_GET_FLAGS(x)	((x & 0x00cf0000)>>16)
#define TCP_GET_WINDOW(x)	(x & 0x0000ffff)
#define TCP_GET_CHECKSUM(x)	((x & 0xffff0000)>>16)
#define TCP_GET_URGENT(x)	(x & 0x0000ffff)

uint8_t pnetbuffer_read_tcp(pnetbuffer_t *nb,
				uint16_t *src_port,
				uint16_t *dest_port,
				uint32_t *seq_num,
				uint32_t *ack_num,
				uint16_t *flags,
				uint16_t *window,
				uint16_t *checksum,
				uint16_t *urgent,
				char *option_data,
				uint16_t *option_len,
				uint16_t *length)
{
	int		rc;
	uint16_t	len;
	uint32_t	u32;
	char		buffer[256];
	rc = pnetbuffer_read_uint32(nb, &u32);
	if (rc) return rc;
	*src_port = TCP_GET_SRC_PORT(u32);
	*dest_port = TCP_GET_DEST_PORT(u32);

	rc = pnetbuffer_read_uint32(nb, seq_num);
	if (rc) return rc;

	rc = pnetbuffer_read_uint32(nb, ack_num);
	if (rc) return rc;

	rc = pnetbuffer_read_uint32(nb, &u32);
	if (rc) return rc;
	len = TCP_GET_DATA_OSET(u32);
	*flags = TCP_GET_FLAGS(u32);
	*window = TCP_GET_WINDOW(u32);
	*length = len;

	rc = pnetbuffer_read_uint32(nb, &u32);
	if (rc) return rc;
	*checksum = TCP_GET_CHECKSUM(u32);
	*urgent = TCP_GET_URGENT(u32);

	if (len > 20) {
		if (option_data) {
			rc = pnetbuffer_read_data(nb, option_data, len-20);
			if (rc) return rc;
			*option_len = len-20;
		} else {
			rc = pnetbuffer_read_data(nb, buffer, len-20);
			if (rc) return rc;
		}
	}

	return 0;
}






