/************************************************************************/
/* backend.h - DNS backend connection management			*/
/************************************************************************/
/*
 *  Copyright 2021-2023 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#ifndef _BACKEND_H_
#define _BACKEND_H_


#include <stdint.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <cutil/cutil.h>

#include <pfm.h>
#include <metric.h>




#define BACKEND_RECONNECT_BACKOFF_BASE		5000
#define BACKEND_RECONNECT_BACKOFF_MAX		300000


#define BACKEND_F_RECONNECT_PENDING		1
#define BACKEND_F_TIMER_PENDING			2


typedef enum {
	BACKEND_PROTO_UDP	= 0,
	BACKEND_PROTO_TCP,
	_BACKEND_PROTO_MAX,
} backend_protocol;


typedef enum {
	METRIC_BE_REQUESTS	= 0,
	_METRIC_BE_COUNT,
} backend_metric_index;


typedef struct _backend_t backend_t;
struct _backend_t {

	backend_protocol	protocol;

	uint32_t		flags;

	char			*dest_host;
	char			*dest_service;
	struct sockaddr		dest_sa;

	int			conn_max;
	int			conn_min;
	int			conn_count;
	int			*conn_fd_table;

	pfm_timer_t		*timer;
	int			timer_backoff_ms;

	metric_t		metric[_METRIC_BE_COUNT];

	// usage rate info (for backend connection scaling)
	uint32_t		rate_last_count;
	time_t			rate_last_time;
};


backend_t *backend_new(backend_protocol protocol,
			char *hostname,
			char *servicename,
			int min_connections,
			int max_connections);
int backend_free(backend_t *be);
int backend_increase_connections(backend_t *be, int count);
int backend_decrease_connections(backend_t *be, int count);
int backend_replace_connection(backend_t *be, int fd);

// forcibly increase by one connection (within maximum)
int backend_add_connection(backend_t *be);

backend_protocol backend_name_to_protocol(char *name);
char *backend_protocol_to_name(backend_protocol protocol);

void backend_count_usage(backend_t *be);

#endif


