/************************************************************************/
/* fdfuncs.h - fd_t manipulation functions				*/
/************************************************************************/
/*
 *  Copyright 2021-2024 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#ifndef _FDFUNCS_H_
#define _FDFUNCS_H_

#include <stdint.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <dlist.h>
#include <pfm_dns.h>


typedef enum {
	FD_CONN_TYPE_UDP	= 0,
	FD_CONN_TYPE_TCP,
	_FD_CONN_TYPE_MAX,
} fd_conn_type;


typedef enum {
	FD_USAGE_TYPE_SERVER	= 0,
	FD_USAGE_TYPE_FRONTEND,
	FD_USAGE_TYPE_BACKEND,
	_FD_USAGE_TYPE_MAX,
} fd_usage_type;




typedef struct _fd_t fd_t;
struct _fd_t {
	// the socket itself
	int		fd;

	fd_conn_type	conntype;
	fd_usage_type	usagetype;

	// backend pointer (only for backend FDs)
	void		*backend;

	// FD state pointer (request_table_t, etc)
	union {
		void			*ptr;
		request_table_t		*request_table;
	} ptr;

	packetbuf_t	*in_pb;	// curr. inbound packet
	dlist_t		in_qhead; // queued inbound, complete packets

	packetbuf_t	*out_pb;  // curr. outbound packet
	dlist_t		out_qhead; // queued outbound packets

	// TCP frontend connection source info
	struct sockaddr	saddr;
	socklen_t	saddr_len;
};



void fd_entry_init(fd_t *fde,
		int fd,
		fd_conn_type conntype,
		fd_usage_type usagetype,
		void *dataptr);
void fd_entry_clear(fd_t *fde);


// attempt to write to/read from the socket
// (using internally managed packetbuf_t queues)
int fd_try_write(fd_t *fde);
int fd_try_read(fd_t *fde);


// submit a packetbuf_t for writing to the socket
int fd_submit_for_write(fd_t *fde, packetbuf_t *pb);

// return a complete packet read from the socket
// (returns NULL if no complete packet is waiting
packetbuf_t *fd_get_ready_packet(fd_t *fde);



// inline debug functions
static inline void dump_fdte(fd_t *fde)
{
	struct sockaddr_in      *sin;

	printf("%s() fdte=%p", __func__, fde);
	if (!fde) return;
	if (fde->conntype == FD_CONN_TYPE_TCP) {
		sin = (struct sockaddr_in *)&fde->saddr;
		printf(
			"%s() .conntype=%d .usagetype=%d .backend=%p .in_pb=%p"
			" .out_pb=%p out_qhead.=%p .in_qhead=%p saddr=%s\n",
			__func__,
			fde->conntype,
			fde->usagetype,
			fde->backend,
			fde->in_pb,
			fde->out_pb,
			&fde->out_qhead,
			&fde->in_qhead,
			inet_ntoa(sin->sin_addr)
		);
	} else {
		printf(
			"%s() .conntype=%d .usagetype=%d .backend=%p .in_pb=%p"
			" .out_pb=%p out_qhead.=%p .in_qhead=%p\n",
			__func__,
			fde->conntype,
			fde->usagetype,
			fde->backend,
			fde->in_pb,
			fde->out_pb,
			&fde->out_qhead,
			&fde->in_qhead
		);
	}
//	dump_packetbuf(fde->in_pb);
//	dump_packetbuf(fde->out_pb);
	return;
}



#endif


